"""redcap tap class."""

from typing import Any, List

from singer_sdk import Tap
from singer_sdk import typing as th  # JSON schema typing helpers

# TODO: Import your custom stream types here:
from tap_redcap.streams import RedCapEnrollmentStream, RedCapFormStream

STREAM_TYPES = [RedCapFormStream, RedCapEnrollmentStream]


class TapRedCap(Tap):
    """redcap tap class."""

    name = "tap-redcap"

    # TODO: Update this section with the actual config values you expect:
    config_jsonschema = th.PropertiesList(
        th.Property(
            "token",
            th.StringType,
            description="The token to authenticate against the API service",
        ),
        th.Property(
            "base_url", th.StringType, description="Base URL for RedCap API service"
        ),
        th.Property(
            "api_timeout", th.IntegerType, description="Base URL for RedCap API service"
        ),
        th.Property("action", th.StringType, description="Specify form for export."),
        th.Property("content", th.StringType, description="Specify form for export."),
        th.Property(
            "forms", th.ArrayType(th.StringType), description="Specify form for export."
        ),
        th.Property("events", th.StringType, description="Specify events for export."),
        th.Property("format", th.StringType, description="Specify format for export."),
        th.Property(
            "rawOrLabel",
            th.StringType,
            description="Specify raw or label data for export.",
        ),
        th.Property(
            "rawOrLabelHeaders",
            th.StringType,
            description="Specify raw or label data headers for export.",
        ),
        th.Property(
            "exportCheckboxLabel",
            th.StringType,
            description="Specify checbox labels for export.",
        ),
        th.Property(
            "exportSurveyFields",
            th.StringType,
            description="Specify survey fields for export.",
        ),
        th.Property(
            "returnFormat",
            th.StringType,
            description="Specify survey return format for export.",
        ),
        th.Property("type", th.StringType, description="Specify type for export."),
        th.Property(
            "exportDataAccessGroups",
            th.StringType,
            description="Specify export for data access group.",
        ),
    ).to_dict()

    def discover_streams(self) -> List[Any]:
        """Return a list of discovered streams."""
        forms: List[str] = self.config["forms"]
        stream_list = []
        for form in forms:
            if form == "enrollment":
                return [RedCapEnrollmentStream(self, form)]
            stream_list.append(RedCapFormStream(self, form))
        return stream_list



"""REST client handling, including redcapStream base class."""

from pathlib import Path
from typing import Any, Dict, Iterable, Optional, TypeVar, cast

import requests
from singer_sdk.helpers.jsonpath import extract_jsonpath
from singer_sdk.helpers.types import Context
from singer_sdk.streams import RESTStream

SCHEMAS_DIR = Path(__file__).parent / Path("./schemas")


class RedCapStream(RESTStream):
    """Redcap stream class."""

    _TToken = TypeVar("_TToken")

    # TODO: Set the API's base URL here:
    # url_base = "https://redcap.chop.edu/api/"

    # OR use a dynamic url_base:
    @property
    def url_base(self) -> str:
        """Return the API URL root, configurable via tap settings."""
        return self.config["base_url"]

    @property
    def timeout(self) -> int:
        """Return the API timeout interval, configurable via tap settings."""
        return self.config["api_timeout"]

    records_jsonpath = "$[*]"  # Or override `parse_response`.
    next_page_token_jsonpath = "$.next_page"  # Or override `get_next_page_token`.

    @property
    def http_headers(self) -> dict:
        """Return the http headers needed."""
        headers = {}
        if "user_agent" in self.config:
            headers["User-Agent"] = self.config.get("user_agent")
        # If not using an authenticator, you may also provide inline auth headers:
        # headers["Private-Token"] = self.config.get("auth_token")
        return headers

    def get_next_page_token(
        self, response: requests.Response, previous_token: Optional[Any]
    ) -> Optional[Any]:
        """Return a token for identifying next page or None if no more pages."""
        # TODO: If pagination is required, return a token which can be used to get the
        #       next page. If this is the final page, return "None" to end the
        #       pagination loop.
        if self.next_page_token_jsonpath:
            all_matches = extract_jsonpath(
                self.next_page_token_jsonpath, response.json()
            )
            first_match = next(iter(all_matches), None)
            next_page_token = first_match
        else:
            next_page_token = response.headers.get("X-Next-Page", None)

        return next_page_token

    def get_url_params(
        self, context: Optional[Context], next_page_token: Optional[_TToken]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization."""
        params: dict = {}
        if next_page_token:
            params["page"] = next_page_token
        if self.replication_key:
            params["sort"] = "asc"
            params["order_by"] = self.replication_key
        return params

    def parse_response(self, response: requests.Response) -> Iterable[dict]:
        """Parse the response and return an iterator of result rows."""
        # TODO: Parse response body and return a set of records.
        yield from extract_jsonpath(self.records_jsonpath, input=response.json())

    # overrides RESTStream::prepare_request_payload
    def prepare_request_payload(
        self, context: Optional[Context], next_page_token: Optional[_TToken]
    ) -> Optional[dict]:
        """Create the request payload from the object."""
        return {
            "token": self.config["token"],
            "content": self.config["content"],
            "action": self.config["action"],
            "format": self.config["format"],
            "type": self.config["type"],
            "forms": "",  # Not a valid attribute
            "events": self.config["events"],
            "rawOrLabel": self.config["rawOrLabel"],
            "rawOrLabelHeaders": self.config["rawOrLabelHeaders"],
            "exportCheckboxLabel": self.config["exportCheckboxLabel"],
            "exportSurveyFields": self.config["exportSurveyFields"],
            "returnFormat": self.config["returnFormat"],
            "exportDataAccessGroups": self.config["exportDataAccessGroups"],
        }

    # overrides RESTStream::prepare_request
    def prepare_request(
        self, context: Optional[Context], next_page_token: Optional[_TToken]
    ) -> requests.PreparedRequest:
        """Create the request for RedCAP."""
        request_data = self.prepare_request_payload(context, next_page_token)

        request = cast(
            requests.PreparedRequest,
            self.requests_session.prepare_request(
                requests.Request(
                    method="POST",
                    url=self.url_base,
                    data=request_data,
                ),
            ),
        )
        return request



"""Stream type classes for tap-redcap."""

from singer_sdk import Tap
from singer_sdk import typing as th  # JSON Schema typing helpers

from tap_redcap.client import RedCapStream


class RedCapFormStream(RedCapStream):
    """A streaming client of RedCAP survey form results."""

    def __init__(self, tap: Tap, form: str):
        """Initialize for survey form streaming client."""
        super().__init__(tap, f"stream_{form}")
        self.forms = form

    path = ""
    schema = th.PropertiesList(
        th.Property("record", th.StringType, description="The RedCap record ID"),
        th.Property(
            "redcap_event_name", th.StringType, description="RedCap event name"
        ),
        th.Property("field_name", th.StringType, description="RedCap field name"),
        th.Property("value", th.StringType, description="Field value entered by user"),
        th.Property(
            "redcap_repeat_instrument",
            th.StringType,
            description="RedCap repeating instrument name",
        ),
        th.Property(
            "redcap_repeat_instance",
            th.NumberType,
            description="RedCap repeat instance",
        ),
    ).to_dict()


class RedCapEnrollmentStream(RedCapStream):
    """A streaming client of RedCAP enrollment data."""

    def __init__(self, tap: Tap, form: str):
        """Create new instance of RedCAP enrollment streaming client."""
        super().__init__(tap, f"stream_{form}")
        self.forms = form

    path = ""
    schema = th.PropertiesList(
        th.Property("study_id", th.StringType, description="The RedCap record ID"),
        th.Property(
            "redcap_event_name", th.StringType, description="RedCap event name"
        ),
        th.Property(
            "redcap_repeat_instrument",
            th.StringType,
            description="RedCap repeating instrument name",
        ),
        th.Property(
            "redcap_repeat_instance",
            th.NumberType,
            description="RedCap repeat instance",
        ),
        th.Property(
            "redcap_data_access_group",
            th.StringType,
            description="The RedCap Data Access Group",
        ),
        th.Property(
            "subject_consent_date",
            th.StringType,
            description="The Subject Consent Date",
        ),
        th.Property(
            "enrollment_status",
            th.StringType,
            description="The Subject Enrollment Status",
        ),
        th.Property("phi", th.StringType, description="The Subject Enrollment Status"),
        th.Property("first_name", th.StringType, description="The Subject First Name"),
        th.Property("last_name", th.StringType, description="The Subject Last Name"),
        th.Property(
            "mrn", th.StringType, description="The Subject Medical Record Number"
        ),
        th.Property("dob", th.StringType, description="The Subject Date of Birth"),
        th.Property(
            "enrollment_complete",
            th.StringType,
            description="Completion status of the Enrollment Form",
        ),
    ).to_dict()



"""Tests standard tap features using the built-in SDK tests library."""

from os import getenv

from singer_sdk.testing import get_standard_tap_tests

from tap_redcap.tap import TapRedCap

REDCAP_API_TOKEN = getenv("REDCAP_API_TOKEN", "")

SAMPLE_CONFIG = {
    "token": REDCAP_API_TOKEN,
    "content": "record",
    "action": "export",
    "format": "json",
    "type": "eav",
    "forms": ["demographics_form"],
    "events": "",
    "rawOrLabel": "label",
    "rawOrLabelHeaders": "label",
    "exportCheckboxLabel": "true",
    "exportSurveyFields": "true",
    "returnFormat": "json",
    "exportDataAccessGroups": "true",
    "base_url": "https://redcap.chop.edu/api/",
    "api_timeout": 1500,
    # TODO: Initialize minimal tap config
}


# Run standard built-in tap tests from the SDK:
def test_standard_tap_tests():
    """Run standard tap tests from the SDK."""
    tests = get_standard_tap_tests(TapRedCap, config=SAMPLE_CONFIG)
    for test in tests:
        test()


# TODO: Create additional tests as appropriate for your tap.
